/*

        Python wrapper for aspell
        
        Tested with:
        * aspell 0.50.5 & python 2.1
				* aspell 0.50.5 & python 2.3.4
        
        Released under GNU GPL license
        
        Wojciech Mua
        wmula[at]o2[dot]pl

 History:
 # 20-22.08.2004:
              * first version of module
 #    28.08.2004:
              * tested with python 2.3.4
              * now aspell.new accepts
                (see typescript2.txt for examples)
 #     7.10.2004:
              * fixed saveAllwords method
                patch by Helmut Jarausch

 ============================================================================
 
 The module aspell provides a type AspellSpeller and function new() that
 creating AspellSpeller object.
 Function new() can get a list of tuples (string, string) - the first
 string is a config key, second its value.

 Each AspellSpeller object has its own configuration. The detailed list 
 of all configuration options is in the 4-th section of manual "Customizing
 aspell".

 AspellSpeller object has following methods:

 * check(word)
        Checks spelling of word. Returns 1 if word is correct, 0 otherwise.

 * suggest(word)
        Returns a list of suggested spellings for given word. Even if word 
        is correct (i.e. check(word) returned 1) aspell performs action.

 * getMainwordlist()
 * getPersonalwordlist()
 * getSessionwordlist()
        These functions return a lists of words stored in main, personal
        and session dictionary.
 
 * addtoPersonal(word)
 * addtoSession(word)
        These functions add a word to personal and session dictionary.

 * saveAllwords()
        Saves all words added thru addtoPersonal() and addtoSession()
        methods.

 * clearSession()
        Clear current session, i.e. erases all words added thru
        addtoSession method since last saveallwords() call.

 * addReplacement(misspelled word, correct word)
        Add a replacement pair, i.e. a misspeled and correct words.
        For example 'teh' and 'the'.

 * setoption(key, value)
        Sets single option for speller. Returns None on succes, raises
        exception when key does not exist, value is invalid, or aspell
        failed.

 * getoption(key)
 * getdefoption(key)
        Returns current or default value of given option key.
        Raises exception when key does not exist.

 * getconfigkeys()
        Returns list of all options. List item is a 4-tuple:
             1. key name
             2. key type = string|integer|boolean|list
             3. default value
             4. short description or "internal"

 */

#include <Python.h>
#include <aspell.h>

staticforward PyTypeObject aspell_AspellType;
static PyObject* AspellException;

typedef struct {
	PyObject_HEAD
	AspellSpeller* speller;	/* the speller       */
	AspellConfig*  config;	/* spellers's config */
} aspell_AspellObject;

#define Config(pyobject)  (((aspell_AspellObject*)pyobject)->config)
#define Speller(pyobject) (((aspell_AspellObject*)pyobject)->speller)

#define return_None   \
	Py_INCREF(Py_None); \
	return Py_None;

/* Create a new speller *******************************************************/
static PyObject* aspell_new(PyObject* self, PyObject* args) {
	aspell_AspellObject* newobj;
	
	AspellSpeller* speller = 0;
	AspellConfig*  config;
	AspellCanHaveError* possible_error;

	PyObject *list = NULL; /* optional argument */
	PyObject *item;
	int i, len;
	char *key, *value;
	
	config = new_aspell_config();
	if (config == NULL) {
		PyErr_SetString(AspellException, "Can't create config");
		return NULL;
	}

	/* check constructor arguments */
	if (!PyArg_ParseTuple(args, "|O:new_aspell", &list))
		return NULL;

	/* optional arg present */
	if (list != NULL) {
		/* check if it is list */
		if (!PyList_CheckExact(list)) {
			PyErr_SetString(PyExc_TypeError, "Expeced list");
			delete_aspell_config(config);
			return NULL;
		}
		/* read list; each item must be a tuple of two strings */
		len = PyList_Size(list);
		for (i=0; i<len; i++) {
			item = PyList_GetItem(list, i); /* item = list[i]; refcount preserved */
			if (!PyTuple_CheckExact(item)) {
				PyErr_SetString(PyExc_TypeError, "List item must be a tuple (string, string).");
				return NULL;
			}
			if (!PyArg_ParseTuple(item, "ss", &key, &value)) {
				PyErr_SetString(PyExc_TypeError, "List item must be a tuple (string, string).");
				delete_aspell_config(config);
				return NULL;
			}

			if (!aspell_config_replace(config, key, value)) {
				PyErr_SetString(AspellException, aspell_config_error_message(config));
				delete_aspell_config(config);
				return NULL;
			}
		}
	}
	
	/* try to create a new speller */
	possible_error = new_aspell_speller(config);
	delete_aspell_config(config);

	if (aspell_error_number(possible_error) == 0)
		/* save a speller */
		speller = to_aspell_speller(possible_error);
	else {
		/* or raise an exception */
		PyErr_SetString(AspellException, aspell_error_message(possible_error));
		delete_aspell_can_have_error(possible_error);
		return NULL;
	}

	/* create a new py-object */
  newobj = (aspell_AspellObject*)PyObject_New(aspell_AspellObject, &aspell_AspellType);
	newobj->speller = speller;
	newobj->config  = aspell_speller_config(speller);

	return (PyObject*)newobj;
}

/* Delete speller *************************************************************/
static void aspell_aspell_dealloc(PyObject* self) {
	delete_aspell_speller( Speller(self) );
	PyObject_Del(self);
}

/* method:check ***************************************************************/
static PyObject* m_check(PyObject* self, PyObject* args) {
	char* word;
	int   length;

	if (!PyArg_ParseTuple(args, "s#", &word, &length)) {
		PyErr_SetString(PyExc_TypeError, "Bad argument type.");
		return NULL;
	}

	if (!length) {
		PyErr_SetString(PyExc_ValueError, "Empty string passed");
		return NULL;
	}

	switch (aspell_speller_check(Speller(self), word, length)) {
		case 0:
			return Py_BuildValue("i", 0);
		case 1:
			return Py_BuildValue("i", 1);
		default:
			PyErr_SetString(AspellException, aspell_speller_error_message(Speller(self)));
			return NULL;
	}
}

/* helper function: converts an aspell list into python list */
static PyObject* aspelllist2PythonList(const AspellWordList* wordlist) {
	PyObject* list;
	AspellStringEnumeration* elements;
	const char* word;

	list = PyList_New(0);
	if (!list) {
		PyErr_SetString(PyExc_Exception, "Can't create new list.");
		return NULL;
	}

	elements = aspell_word_list_elements(wordlist);
	while ( (word=aspell_string_enumeration_next(elements)) != 0)
		if (PyList_Append(list, Py_BuildValue("s", word)) == -1) {
			PyErr_SetString(PyExc_Exception, "Can't append.");
			delete_aspell_string_enumeration(elements);
			return NULL;
		}
	delete_aspell_string_enumeration(elements);
	return list;
}

/* method:suggest ************************************************************/
static PyObject* m_suggest(PyObject* self, PyObject* args) {
	char* word;
	int   length;

	if (!PyArg_ParseTuple(args, "s#", &word, &length)) {
		PyErr_SetString(PyExc_TypeError, "Bad argument type");
		return NULL;
	}
	if (!length) {
		PyErr_SetString(PyExc_ValueError, "Empty string passed");
		return NULL;
	}
	return aspelllist2PythonList( aspell_speller_suggest(Speller(self), word, length));
}

/* method:getconfigkeys *******************************************************/
static PyObject* m_getconfigkeys(PyObject* self, PyObject* args) {
	AspellKeyInfoEnumeration *keys_enumeration;
	const AspellKeyInfo *key_info;
	
	PyObject *key_list;
	char *key_type = 0;

	keys_enumeration = aspell_config_possible_elements(Config(self), 1);
	if (!keys_enumeration) {
		PyErr_SetString(AspellException, "Couldn't get list of keys (aspell returned NULL)");
		return NULL;
	}

	key_list = PyList_New(0);
	while ((key_info = aspell_key_info_enumeration_next(keys_enumeration))) {
		switch (key_info->type) {
			case AspellKeyInfoString:
				key_type = "string";
				break;
			case AspellKeyInfoInt:
				key_type = "integer";
				break;
			case AspellKeyInfoBool:
				key_type = "boolean";
				break;
			case AspellKeyInfoList:
				key_type = "list";
				break;
		}
		if (PyList_Append(key_list, Py_BuildValue("(ssss)", key_info->name, key_type, key_info->def, key_info->desc ? key_info->desc : "internal")) == -1) {
			PyErr_SetString(PyExc_Exception, "Can't append.");
			delete_aspell_key_info_enumeration(keys_enumeration);
			return NULL;
		}
	}
	delete_aspell_key_info_enumeration(keys_enumeration);
	return key_list;
}

/* method:getoption ***********************************************************/
static PyObject* m_getoption(PyObject* self, PyObject* args) {
	char *opt;
	const char* val;

	if (!PyArg_ParseTuple(args, "s", &opt)) {
		PyErr_SetString(PyExc_TypeError, "Bad argument type");
		return NULL;
	}

	val = aspell_config_retrieve(Config(self), opt);
	if (aspell_config_error(Config(self)) != NULL) {
		PyErr_SetString(AspellException, aspell_config_error_message(Config(self)));
		return NULL;
	}
	else
		return Py_BuildValue("s", val);
}

/* method:getdefoption ********************************************************/
static PyObject* m_getdefoption(PyObject* self, PyObject* args) {
	char *opt;
	const char* val;

	if (!PyArg_ParseTuple(args, "s", &opt)) {
		PyErr_SetString(PyExc_TypeError, "Bad argument type.");
		return NULL;
	}

	val = aspell_config_get_default(Config(self), opt);
	if (aspell_config_error(Config(self)) != NULL) {
		PyErr_SetString(AspellException, aspell_config_error_message(Config(self)));
		return NULL;
	}
	else
		return Py_BuildValue("s", val);
}


/* method:setoption ***********************************************************/
static PyObject* m_setoption(PyObject* self, PyObject* args) {
	char *opt, *val;

	if (!PyArg_ParseTuple(args, "ss", &opt, &val)) {
		PyErr_SetString(PyExc_TypeError, "Invalid argument");
		return NULL;
	}

	if (aspell_config_replace(Config(self), opt, val)) {
		return_None;
	}
	else {
		PyErr_SetString(AspellException, aspell_config_error_message(Config(self)));
		return NULL;
	}
}

/* method:getMainwordlist *****************************************************/
static PyObject* m_getMainwordlist(PyObject* self, PyObject* args) {
	return aspelllist2PythonList( aspell_speller_main_word_list(Speller(self)));
}

/* method:getPersonalwordlist *************************************************/
static PyObject* m_getPersonalwordlist(PyObject* self, PyObject* args) {
	return aspelllist2PythonList( aspell_speller_personal_word_list(Speller(self)));
}

/* method:getSessionwordlist **************************************************/
static PyObject* m_getSessionwordlist(PyObject* self, PyObject* args) {
	return aspelllist2PythonList( aspell_speller_session_word_list(Speller(self)));
}

/* method:addtoPersonal *******************************************************/
static PyObject* m_addtoPersonal(PyObject* self, PyObject* args) {
	char *word;
	int   length;
	
	if (!PyArg_ParseTuple(args, "s#", &word, &length)) {
		PyErr_SetString(PyExc_TypeError, "Invalid argument");
		return NULL;
	}

	aspell_speller_add_to_personal(Speller(self), word, length);
	if (aspell_speller_error(Speller(self)) != 0) {
		PyErr_SetString(AspellException, aspell_speller_error_message(Speller(self)));
		return NULL;
	}
	return_None;
}

/* method:addtoSession ********************************************************/
static PyObject* m_addtoSession(PyObject* self, PyObject* args) {
	char *word;
	int   length;
	
	if (!PyArg_ParseTuple(args, "s#", &word, &length)) {
		PyErr_SetString(PyExc_TypeError, "Invalid argument");
		return NULL;
	}

	aspell_speller_add_to_session(Speller(self), word, length);
	if (aspell_speller_error(Speller(self)) != 0) {
		PyErr_SetString(AspellException, aspell_speller_error_message(Speller(self)));
		return NULL;
	}
	return_None;
}

/* method:clearsession ********************************************************/
static PyObject* m_clearsession(PyObject* self, PyObject* args) {
	aspell_speller_clear_session(Speller(self));
	if (aspell_speller_error(Speller(self)) != 0) {
		PyErr_SetString(AspellException, aspell_speller_error_message(Speller(self)));
		return NULL;
	}

	return_None;
}

/* method:saveallwords ********************************************************/
static PyObject* m_saveallwords(PyObject* self, PyObject* args) {
	aspell_speller_save_all_word_lists(Speller(self));
	if (aspell_speller_error(Speller(self)) != 0) {
		PyErr_SetString(AspellException, aspell_speller_error_message(Speller(self)));
		return NULL;
	}
	return_None;
}
/* method:addReplacement ******************************************************/
static PyObject* m_addReplacement(PyObject* self, PyObject* args) {
	char *mis; int ml;
	char *cor; int cl;
	
	if (!PyArg_ParseTuple(args, "s#s#", &mis, &ml, &cor, &cl)) {
		PyErr_SetString(PyExc_TypeError, "Invalid arguments");
		return NULL;
	}
	aspell_speller_store_replacement(Speller(self), mis, ml, cor, cl);
	if (aspell_speller_error(Speller(self)) != 0) {
		PyErr_SetString(AspellException, aspell_speller_error_message(Speller(self)));
		return NULL;
	}
	return_None;
}

/* AspellSpeller methods table */
static PyMethodDef aspell_object_methods[] = {
	{
		"check",
		(PyCFunction)m_check,
		METH_VARARGS,
		"check(word) => 0 or 1\n"
 		"Checks spelling of word.\n"
		"Returns 1 if word is correct, 0 otherwise."
	},
	{
		"suggest",
		(PyCFunction)m_suggest,
		METH_VARARGS,
		"suggest(word) => list of words\n"
 		"Returns a list of suggested spelling for given word.\n"
		"Even if word is correct (i.e. check(word) returned 1) aspell performs action."
	},
	{
		"setoption",
		(PyCFunction)m_setoption,
		METH_VARARGS,
		"setoption(key, value) => None\n"
		"Sets single option for speller.\n"
		"Returns None on succes, raises exception when key does not exist, value is invalid, or aspell failed."
	},
	{
		"getoption",
		(PyCFunction)m_getoption,
		METH_VARARGS,
		"getoption(key) => value\n"
 		"Returns current value of given option key.\n"
		"Raises exception when key does not exist."
	},
	{
		"getdefoption",
		(PyCFunction)m_getdefoption,
		METH_VARARGS,
		"getdefoption(key) => value\n"
 		"Returns default value of given option key.\n"
		"Raises exception when key does not exist."
	},
	{
		"getconfigkeys",
		(PyCFunction)m_getconfigkeys,
		METH_VARARGS,
		"getconfigkeys(key) => value\n"
		"Fetch all config keys. The list item is a 4-touple:\n"
		"\t1. key name\n"
		"\t2. key type={string|integer|boolean|list}\n"
		"\t3. default value\n"
		"\t4. short description."
	},
	{
		"getMainwordlist",
		(PyCFunction)m_getMainwordlist,
		METH_VARARGS,
		"getMainwordlist() => list of words\n"
		"Return a list of words stored in the main dictionary."

	},
	{
		"getPersonalwordlist",
		(PyCFunction)m_getPersonalwordlist,
		METH_VARARGS,
		"getPersonalwordlist() => list of words\n"
		"Return a list of words stored in the personal dictionary."
	},
	{
		"getSessionwordlist",
		(PyCFunction)m_getSessionwordlist,
		METH_VARARGS,
		"getSessionwordlist() => list of words\n"
		"Return a list of words stored in the session dictionary."
	},
	{
		"clearSession",
		(PyCFunction)m_clearsession,
		METH_VARARGS,
		"clearSession() => None\n"
		"Clear current session, i.e. erases all words added thru addtoSession method since last saveallwords() call."
	},
	{
		"saveAllwords",
		(PyCFunction)m_saveallwords,
		METH_VARARGS,
		"saveAllwords() => None\n"
 		"Save all words added thru addtoPersonal() and addtoSession() methods."
	},
	{
		"addtoPersonal",
		(PyCFunction)m_addtoPersonal,
		METH_VARARGS,
		"addtoPersonal(word) => None\n"
		"Add word to the personal dictionary"
	},
	{
		"addtoSession",
		(PyCFunction)m_addtoSession,
		METH_VARARGS,
		"addtoSession(word) => None\n"
		"Add word to the session dictionary"
	},
	{
		"addReplacement",
		(PyCFunction)m_addReplacement,
		METH_VARARGS,
		"addReplacement(misspeled word, correct word) => None\n"
		"Add a replacement pair, i.e. a misspeled and correct words.\n"
		"For example 'teh' and 'the'."
	},
	{NULL, NULL, 0, NULL}
};

static PyObject* aspell_aspell_getattr(PyObject *obj, char *name) {
	return Py_FindMethod(aspell_object_methods, obj, name);
}

static PyTypeObject aspell_AspellType = {
	PyObject_HEAD_INIT(NULL)
	0,
	"AspellSpeller",
	sizeof(aspell_AspellObject),
	0,
	
	aspell_aspell_dealloc,
	0,
	aspell_aspell_getattr
};

static PyMethodDef aspell_methods[] = {
	{"new", aspell_new, METH_VARARGS, "Create a new AspellSpeller"},
	{NULL, NULL, 0, NULL}
};

DL_EXPORT(void)
initaspell(void) {
	PyObject *module;
	PyObject *dict;
	
	aspell_AspellType.ob_type = &PyType_Type;
	module = Py_InitModule("aspell", aspell_methods);
	dict   = PyModule_GetDict(module);
	AspellException = PyErr_NewException("aspell.error", NULL, NULL);
	PyDict_SetItemString(dict, "error", AspellException);
}

/*
vim:ts=2
*/
